# -*- coding: utf-8 -*-
import time
import uuid
from typing import Any, Awaitable, Callable, Optional, Type, Union

from pydantic import ValidationError

from .context import meta_context_var
from .errors import (
    JarpcError,
    JarpcInvalidRequest,
    JarpcParseError,
    JarpcServerError,
    raise_exception,
)
from .format import (
    JarpcRequest,
    JarpcResponse,
    RequestT,
    ResponseT,
    json_dumps,
    json_loads,
)


class JarpcClient:
    """
    JARPC Client implementation.

    To make RPC it requires transport.
    Transport gets JARPC request as string, JarpcRequest-object and kwargs given with client call.
    If rsvp is True, transport must return JARPC response string, otherwise transport may not return any result.
    Transport's exceptions will be overwritten with `JarpcServerError` unless they are `JarpcError` subclasses.

    Example of usage with python "requests" library:
    ```
    def requests_transport(request_string, request, timeout=60.0):
        try:
            return requests.post(url='https://kitchen.org/jsonrpc', data=request_string, timeout=timeout)
        except Timeout:
            raise JarpcTimeout

    kitchen = JarpcClient(transport=requests_transport)
    salad = kitchen(method='cook_salad', params=dict(name='Caesar'), request_id='1', timeout=15)
    ```

    You can also define transport as class:
    ```
    class RequestsTransport:
        def __init__(self, url, headers=None):
            self.url = url
            self.headers = headers or {}
        def __call__(self, request_string, request, timeout=60.0):
            try:
                return requests.post(url=self.url, headers=self.headers, data=request_string, timeout=timeout)
            except Timeout:
                raise JarpcTimeout

    transport = RequestsTransport(url='https://kitchen.org/jsonrpc', headers={'Content-Type': 'application/json'})
    kitchen = JarpcClient(transport)
    salad = kitchen(method='cook_salad', params=dict(name='Caesar'), request_id='1', timeout=15)
    ```

    If you don't need to pass JARPC meta params and transport kwargs, you can use method-like calling syntax:
    ```
    salad = kitchen.cook_salad(name='Caesar')
    ```
    """

    def __init__(
        self,
        # transport: Callable[[str, JarpcRequest, **kwargs], Union[str, None]]
        transport: Callable[
            [str, JarpcRequest, Any | None], Awaitable[str | None] | str | None
        ],
        default_ttl: Optional[float] = None,
        default_rpc_ttl: Optional[float] = None,
        default_notification_ttl: Optional[float] = None,
    ):
        """
        :param transport: callable to send request
        :param default_ttl: float time interval while calling still actual
        :param default_rpc_ttl: default_ttl for rsvp=True calls (if None default_ttl will be used)
        :param default_notification_ttl: default_ttl for rsvp=False calls (if None default_ttl will be used)
        """
        self._transport = transport
        self._default_rpc_ttl = default_rpc_ttl or default_ttl
        self._default_notification_ttl = default_notification_ttl or default_ttl

    def __getattr__(self, method):
        def wrapped(**params):
            return self.simple_call(method=method, params=params)

        return wrapped

    def simple_call(self, method: str, **params):
        return self(method=method, params=params)

    def __call__(
        self,
        method: str,
        params: RequestT,
        ts: Optional[float] = None,
        ttl: Optional[float] = None,
        request_id: Optional[str] = None,
        rsvp: bool = True,
        durable: bool = False,
        **transport_kwargs
    ) -> JarpcResponse:
        request = self._prepare_request(
            method, params, ts, ttl, request_id, rsvp, durable
        )
        request_string = request.model_dump_json()

        try:
            response_string = self._transport(
                request_string, request, **transport_kwargs
            )
        except JarpcError:
            raise
        except Exception as e:
            raise JarpcServerError(e)

        return self._parse_response(response_string, rsvp)

    def _prepare_request(
        self,
        method: str,
        params: RequestT,
        ts: Optional[float] = None,
        ttl: Optional[float] = None,
        request_id: Optional[str] = None,
        rsvp: bool = True,
        durable: bool = False,
        meta: dict[str, Any] = None,
    ) -> JarpcRequest[RequestT]:
        """Make request."""
        if durable:
            ttl = None
        else:
            default_ttl = (
                self._default_rpc_ttl if rsvp else self._default_notification_ttl
            )
            ttl = default_ttl if ttl is None else ttl

        try:
            request = JarpcRequest[RequestT](
                method=method,
                params=params,
                ts=time.time() if ts is None else ts,
                ttl=ttl,
                id=str(uuid.uuid4()) if request_id is None else request_id,
                rsvp=rsvp,
            )
        except ValidationError as e:
            raise JarpcInvalidRequest(e) from e

        return request

    def _parse_response(
        self,
        response_string: str,
        rsvp: bool,
        generic_response_type: Type[ResponseT] = Any,
    ) -> ResponseT | None:
        """Parse response and either return result or raise JARPC error."""
        if rsvp:
            try:
                response = JarpcResponse[generic_response_type].model_validate_json(
                    response_string
                )
            except ValidationError as e:
                raise JarpcServerError(e) from e
            if response.success:
                return response.result
            else:
                error = response.error
                raise_exception(
                    code=error.get("code"),
                    data=error.get("data"),
                    message=error.get("message"),
                )


class AsyncJarpcClient(JarpcClient):
    """
    Asynchronous JARPC Client implementation.

    To make RPC it requires async transport.
    Transport gets JARPC request as string, JarpcRequest-object and kwargs given with client call.
    If rsvp is True, transport must return JARPC response string, otherwise transport may not return any result.
    Transport's exceptions will be overwritten with `JarpcServerError` unless they are `JarpcError` subclasses.

    Example of usage with python "aiohttp" library:
    ```
    async def aiohttp_transport(request_string, request, timeout=60.0):
        try:
            async with aiohttp.ClientSession() as session:
                response = await session.post(url='https://kitchen.org/jsonrpc', data=request_string, timeout=timeout)
                return await response.text()
        except TimeoutError:
            raise JarpcTimeout

    kitchen = AsyncJarpcClient(transport=aiohttp_transport)
    salad = await kitchen(method='cook_salad', params=dict(name='Caesar'), request_id='1', timeout=15)
    ```

    You can also define transport as class:
    ```
    class AiohttpTransport:
        def __init__(self, session, url, headers=None):
            self.session = session
            self.url = url
            self.headers = headers or {}
        async def __call__(self, request_string, request, timeout=60.0):
            try:
                response = await session.post(url=self.url, headers=self.headers, data=request_string, timeout=timeout)
                return await response.text()
            except TimeoutError:
                raise JarpcTimeout

    async with aiohttp.ClientSession() as session:
        transport = AiohttpTransport(session=session, url='https://kitchen.org/jsonrpc',
                                     headers={'Content-Type': 'application/json'})
        kitchen = AsyncJarpcClient(transport=transport)
        salad = await kitchen(method='cook_salad', params=dict(name='Caesar'), request_id='1', timeout=15)
    ```

    If you don't need to pass JARPC meta params and transport kwargs, you can use method-like calling syntax:
    ```
    salad = await kitchen.cook_salad(name='Caesar')
    ```
    """

    async def __call__(
        self,
        method: str,
        params: RequestT,
        ts: Optional[float] = None,
        ttl: Optional[float] = None,
        request_id: Optional[str] = None,
        rsvp: bool = True,
        durable: bool = False,
        meta: dict[str, Any] = None,
        generic_request_type: Type[RequestT] = Any,
        generic_response_type: Type[ResponseT] = Any,
        **transport_kwargs
    ) -> JarpcResponse[ResponseT]:
        context_meta = meta_context_var.get()
        combined_meta = {**context_meta, **(meta or {})} or None

        request: JarpcRequest[RequestT] = self._prepare_request(
            method, params, ts, ttl, request_id, rsvp, durable, combined_meta
        )
        request_string = request.model_dump_json()

        try:
            response_string = await self._transport(
                request_string, request, **transport_kwargs
            )
        except JarpcError:
            raise
        except Exception as e:
            raise JarpcServerError(e)

        return self._parse_response(response_string, rsvp)



# -*- coding: utf-8 -*-
import pytest

from jarpcdantic import (
    JarpcError,
    JarpcExternalServiceUnavailable,
    JarpcForbidden,
    JarpcUnauthorized,
    JarpcUnknownError,
    JarpcValidationError,
    raise_exception,
)


@pytest.mark.parametrize(
    "code, error_class, data",
    [
        (1000, JarpcUnauthorized, "test1"),
        (1001, JarpcForbidden, "test2"),
        (2000, JarpcValidationError, "test3"),
        (3000, JarpcExternalServiceUnavailable, "test3"),
        (9999, JarpcUnknownError, "test4"),
    ],
)
def test_raise_exception(code, error_class, data):
    with pytest.raises(error_class) as e:
        raise_exception(code=code, data=data)

    assert e.value.data == data
    assert e.value.code == code
    assert e.value.message == error_class.message


@pytest.mark.parametrize(
    "code, error_class, message, data",
    [
        (9999, JarpcUnknownError, "test1", "testst"),
        (9999, JarpcUnknownError, "test2", {}),
        (9999, JarpcUnknownError, "test3", ()),
        (9999, JarpcUnknownError, "test4", None),
    ],
)
def test_raise_exception_unknown_error(code, error_class, data, message):
    with pytest.raises(error_class) as e:
        raise_exception(code=code, data=data, message=message)

    assert e.value.data == data
    assert e.value.code == code
    assert e.value.message == message


def test_as_dict_method():
    exception = JarpcError("test_as_dict_method")
    exception.code = 4444
    exception.message = "test exception"

    assert exception.as_dict() == {
        "code": 4444,
        "data": "test_as_dict_method",
        "message": "test exception",
    }



import inspect
import traceback
from collections.abc import Iterable, Mapping
from types import UnionType
from typing import get_origin, get_args, Union, Type, Any

from pydantic import BaseModel, ValidationError

from jarpcdantic import JarpcParseError


def convert_dict_to_model(dict_data: dict, target_model: Type[BaseModel]):
    return target_model.model_validate(dict_data)


def convert_instance_to_another_model(
    instance: BaseModel, target_model: Type[BaseModel]
):
    return target_model.model_validate(instance, from_attributes=True)


def convert_to_pydantic_model(source: dict | BaseModel, target_model: Type[BaseModel]):
    if isinstance(source, dict):
        return convert_dict_to_model(source, target_model)
    elif isinstance(source, BaseModel):
        return convert_instance_to_another_model(source, target_model)
    else:
        raise JarpcParseError(
            f"Cannot convert {type(source)} to {target_model.__name__}"
        )


def convert_single_value(value: Any, target_type: Type) -> Any:
    if inspect.isclass(target_type) and issubclass(target_type, BaseModel):
        return convert_to_pydantic_model(value, target_type)
    if (target_type is None or target_type is type(None)) and value is None:
        return None
    if isinstance(value, target_type):
        return value
    try:
        return target_type(value)
    except (ValueError, TypeError):
        raise JarpcParseError(f"Cannot convert value {value} to {target_type}")


def convert_iterable(value: Any, target_type: Type) -> Any:
    item_type = get_args(target_type)[0]
    origin = get_origin(target_type)
    iterable = (convert_value_to_type(item, item_type) for item in value)
    return origin(iterable)


def convert_mapping(value: Any, target_type: Type) -> Any:
    key_type, value_type = get_args(target_type)
    return {
        convert_value_to_type(k, key_type): convert_value_to_type(v, value_type)
        for k, v in value.items()
    }


def convert_union(value: Any, target_type: Type) -> Any:
    for arg in get_args(target_type):
        if arg is type(None):
            if value is None:
                return None
        else:
            try:
                return convert_value_to_type(value, arg)
            except (ValidationError, JarpcParseError, TypeError, AttributeError):
                continue
    raise JarpcParseError(f"Cannot convert value {value} to any of {target_type}")


def convert_value_to_type(value: Any, target_type: Type) -> Any:
    """
    Универсальная функция для преобразования значения в указанный тип.
    Поддерживает Pydantic модели, контейнеры с моделями и Union.
    """
    origin = get_origin(target_type)

    if origin is None:
        return convert_single_value(value, target_type)
    elif issubclass(origin, Iterable) and not isinstance(value, str):
        return convert_iterable(value, target_type)
    elif origin in {dict, Mapping}:
        return convert_mapping(value, target_type)
    elif origin in (Union, UnionType):
        return convert_union(value, target_type)
    else:
        raise JarpcParseError(f"Unknown type {target_type}")


def convert_params_to_models(params, method_sig):
    """
    Converts parameters from dict to types specified in the method signature.
    """
    converted_params = {}
    for param_name, param_value in params.items():
        param_type = method_sig.parameters.get(
            param_name, inspect.Parameter.empty
        ).annotation
        if param_type is not inspect.Parameter.empty:
            try:
                converted_params[param_name] = convert_value_to_type(
                    param_value, param_type
                )
            except ValidationError as e:
                raise JarpcParseError(
                    f"Invalid parameter '{param_name}' for type {param_type}: {e}"
                )
        else:
            converted_params[param_name] = param_value
    return converted_params


def process_return_value(return_annotation, result):
    if return_annotation is not inspect.Signature.empty:
        try:
            return convert_value_to_type(result, return_annotation)
        except ValidationError as e:
            raise JarpcParseError(
                f"Failed to process return value {result} to type {return_annotation}: {e}"
            )
    return result



# -*- coding: utf-8 -*-
from .client import AsyncJarpcClient, JarpcClient
from .dispatcher import JarpcDispatcher
from .errors import (
    JarpcError,
    JarpcExternalServiceUnavailable,
    JarpcForbidden,
    JarpcInternalError,
    JarpcInvalidParams,
    JarpcInvalidRequest,
    JarpcMethodNotFound,
    JarpcParseError,
    JarpcServerError,
    JarpcTimeout,
    JarpcUnauthorized,
    JarpcUnknownError,
    JarpcValidationError,
    raise_exception,
)
from .format import JarpcRequest, JarpcResponse
from .manager import AsyncJarpcManager, JarpcManager
from .router import JarpcClientRouter

__all__ = (
    # client
    "AsyncJarpcClient",
    "JarpcClient",
    "JarpcClientRouter",
    # dispatcher
    "JarpcDispatcher",
    # errors
    "JarpcError",
    "JarpcExternalServiceUnavailable",
    "JarpcForbidden",
    "JarpcInternalError",
    "JarpcInvalidParams",
    "JarpcInvalidRequest",
    "JarpcMethodNotFound",
    "JarpcParseError",
    "JarpcServerError",
    "JarpcTimeout",
    "JarpcUnknownError",
    "JarpcUnauthorized",
    "JarpcValidationError",
    "raise_exception",
    # format
    "JarpcRequest",
    "JarpcResponse",
    # manager
    "AsyncJarpcManager",
    "JarpcManager",
)

__version__ = "0.1.10"



# -*- coding: utf-8 -*-
import asyncio
import json
import logging
from copy import deepcopy
from datetime import datetime, timezone

import pytest

from jarpcdantic import AsyncJarpcManager, JarpcDispatcher, JarpcManager, JarpcRequest
from jarpcdantic.manager import check_function_call


class TestCheckFunctionCall:
    @pytest.mark.parametrize("context", [{}, {"app": "some app", "db": "db"}])
    @pytest.mark.parametrize(
        "func, params, expected",
        [
            (lambda a, b: ..., dict(a=1, b=2), (True, None)),
            (
                lambda a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, c=3, x=0, y=1),
                (True, None),
            ),
            (lambda a, b, c=0, *, x, y=1, **kw: ..., dict(a=1, b=2, x=0), (True, None)),
            (
                lambda a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0, foo="bar"),
                (True, None),
            ),
            (lambda a, b: ..., dict(a=1, b=2, c=3), (False, "Unexpected arguments: c")),
            (lambda a, b: ..., dict(a=1), (False, "Missing arguments: b")),
            (lambda a, b: ..., dict(), (False, "Missing arguments: a, b")),
            (
                lambda a, b, c=0, *, x, y=1: ...,
                dict(a=1, b=2, c=3, x=0, y=1, z=2),
                (False, "Unexpected arguments: z"),
            ),
            (
                lambda a, b, c=0, *, x, y=1: ...,
                dict(a=1, x=0),
                (False, "Missing arguments: b"),
            ),
            (
                lambda a, b, c=0, *, x, y=1: ...,
                dict(a=1, b=2),
                (False, "Missing arguments: x"),
            ),
        ],
    )
    def test_no_context_effect(self, func, params, context, expected):
        assert check_function_call(func, params, context) == expected

    @pytest.mark.parametrize(
        "func, params, expected",
        [
            (lambda jarpc_request, app, a, b: ..., dict(a=1, b=2), (True, None)),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, c=3, x=0, y=1),
                (True, None),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0),
                (True, None),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0, foo="bar"),
                (True, None),
            ),
            (
                lambda jarpc_request, app, a, b: ...,
                dict(a=1, b=2, c=3),
                (False, "Unexpected arguments: c"),
            ),
            (
                lambda jarpc_request, app, a, b: ...,
                dict(a=1),
                (False, "Missing arguments: b"),
            ),
            (
                lambda jarpc_request, app, a, b: ...,
                dict(),
                (False, "Missing arguments: a, b"),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1: ...,
                dict(a=1, b=2, c=3, x=0, y=1, z=2),
                (False, "Unexpected arguments: z"),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1: ...,
                dict(a=1, x=0),
                (False, "Missing arguments: b"),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1: ...,
                dict(a=1, b=2),
                (False, "Missing arguments: x"),
            ),
            (lambda a, b, app="default", *, x: ..., dict(a=1, b=2, x=0), (True, None)),
            (
                lambda a, b, app="default", *, x: ...,
                dict(a=1, b=2, x=0, app="evil app"),
                (False, "Unexpected arguments: app"),
            ),
            (
                lambda jarpc_request, a, b, *, x: ...,
                dict(a=1, b=2, x=0, jarpc_request="evil request"),
                (False, "Unexpected arguments: jarpc_request"),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, c=3, x=0, y=1),
                (True, None),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0),
                (True, None),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0, foo="bar"),
                (True, None),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0, jarpc_request="evil"),
                (False, "Unavailable arguments: jarpc_request"),
            ),
            (
                lambda jarpc_request, app, a, b, c=0, *, x, y=1, **kw: ...,
                dict(a=1, b=2, x=0, app="evil"),
                (False, "Unavailable arguments: app"),
            ),
        ],
    )
    def test_with_context(self, func, params, expected):
        assert (
            check_function_call(func, params, {"app": "some app", "db": "db"})
            == expected
        )

    @pytest.mark.parametrize(
        "params, expected",
        [
            (dict(a=1, b=2), (True, None)),
            (dict(a=1, b=2, c=3), (False, "Unexpected arguments: c")),
            (dict(self=0, a=1, b=2), (False, "Unexpected arguments: self")),
            (dict(a=1), (False, "Missing arguments: b")),
            (dict(), (False, "Missing arguments: a, b")),
        ],
    )
    def test_method(self, params, expected):
        class SomeMethod:
            def run(self, app, a, b): ...

        assert (
            check_function_call(
                SomeMethod().run, params, {"app": "some app", "db": "db"}
            )
            == expected
        )

    @pytest.mark.parametrize(
        "params, expected",
        [
            (dict(a=1, b=2), (True, None)),
            (dict(a=1, b=2, c=3), (False, "Unexpected arguments: c")),
            (dict(a=1), (False, "Missing arguments: b")),
            (dict(), (False, "Missing arguments: a, b")),
        ],
    )
    def test_static_method(self, params, expected):
        class SomeMethod:
            @staticmethod
            def run(app, a, b): ...

        assert (
            check_function_call(
                SomeMethod().run, params, {"app": "some app", "db": "db"}
            )
            == expected
        )

    @pytest.mark.parametrize(
        "params, expected",
        [
            (dict(a=1, b=2), (True, None)),
            (dict(a=1, b=2, c=3), (False, "Unexpected arguments: c")),
            (dict(self=0, a=1, b=2), (False, "Unexpected arguments: self")),
            (dict(a=1), (False, "Missing arguments: b")),
            (dict(), (False, "Missing arguments: a, b")),
        ],
    )
    def test_object_no_context(self, params, expected):
        class SomeMethod:
            def __call__(self, a, b): ...

        assert check_function_call(SomeMethod(), params, {}) == expected

    @pytest.mark.parametrize(
        "params, expected",
        [
            (dict(a=1, b=2, c=3, x=0, y=1), (True, None)),
            (dict(a=1, b=2, x=0), (True, None)),
            (dict(a=1, b=2, x=0, foo="bar"), (False, "Unexpected arguments: foo")),
            (dict(a=1, b=2, c=3, x=0, y=1, z=2), (False, "Unexpected arguments: z")),
            (dict(self=0, a=1, b=2, x=0), (False, "Unexpected arguments: self")),
            (dict(a=1, x=0), (False, "Missing arguments: b")),
            (dict(a=1, b=2), (False, "Missing arguments: x")),
        ],
    )
    def test_object_with_context(self, params, expected):
        class SomeMethod:
            def __call__(self, jarpc_request, app, a, b, c=0, *, x, y=1): ...

        assert (
            check_function_call(SomeMethod(), params, {"app": "some app", "db": "db"})
            == expected
        )


@pytest.mark.asyncio
class TestGetResponse:
    basic_request = {
        "method": "method",
        "params": {"param": "value"},
        "request_id": "1",
        "version": "1.0",
        "ts": float(1 << 31),
        "ttl": 1.0,
        "rsvp": True,
    }

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_no_method(self, is_async):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )

        if is_async:
            response = await manager.get_response(json.dumps(self.basic_request))
        else:
            response = manager.get_response(json.dumps(self.basic_request))
        assert response is not None
        assert response.result is None
        assert response.error == {
            "code": -32601,
            "message": "Method not found",
            "data": "KeyError: 'method'",
        }

    @pytest.mark.parametrize("is_async", [False, True])
    @pytest.mark.parametrize(
        "params, explanation",
        [
            (dict(a=1, b=2, c=3, x=0, y=1, z=2), "Unexpected arguments: z"),
            (dict(a=1, x=0), "Missing arguments: b"),
            (dict(a=1, b=2), "Missing arguments: x"),
        ],
    )
    async def test_invalid_params(self, is_async, params, explanation):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )
        dispatcher.add_rpc_method(lambda a, b, c=0, *, x, y=1: ..., "method")

        request = deepcopy(self.basic_request)
        request["params"] = params
        if is_async:
            response = await manager.get_response(json.dumps(request))
        else:
            response = manager.get_response(json.dumps(request))
        assert response is not None
        assert response.result is None
        assert response.error == {
            "code": -32602,
            "message": "Invalid params",
            "data": explanation,
        }

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_handle_result_ok(self, is_async):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )

        @dispatcher.rpc_method
        def method(jarpc_request, param):
            return "42"

        if is_async:
            response = await manager.handle(json.dumps(self.basic_request))
        else:
            response = manager.handle(json.dumps(self.basic_request))

        assert json.loads(response)["result"] == "42"

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_handle_result_none(self, is_async):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )

        @dispatcher.rpc_method
        def method(jarpc_request, param):
            return None

        request_data = deepcopy(self.basic_request)
        request_data["ts"] = 1.0

        if is_async:
            response = await manager.handle(json.dumps(request_data))
        else:
            response = manager.handle(json.dumps(request_data))

        assert response is None

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_invalid_request_json_parse_error(self, is_async):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )
        request = "some invalid json"
        if is_async:
            response = await manager.get_response(request)
        else:
            response = manager.get_response(request)
        error = response.error
        assert error["message"] == "Parse error"

    # TODO: Fix test
    # @pytest.mark.parametrize("is_async", [False, True])
    # async def test_invalid_request_no_params(self, is_async):
    #     dispatcher = JarpcDispatcher()
    #     manager = (
    #         AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
    #     )
    #     request = deepcopy(self.basic_request)
    #     request.pop("params")
    #     if is_async:
    #         response = await manager.get_response(json.dumps(request))
    #     else:
    #         response = manager.get_response(json.dumps(request))
    #     error = response.error
    #     assert error["message"] == "Invalid Request"
    #     assert error["data"] == 'Missing required field "params"'

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_invalid_rpc_method(self, is_async):
        """Test rpc_method TypeError"""
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )

        @dispatcher.rpc_method
        def method(jarpc_request, param):
            return "2" + 2

        if is_async:
            response = await manager.get_response(json.dumps(self.basic_request))
        else:
            response = manager.get_response(json.dumps(self.basic_request))
        error = response.error
        assert error["message"] == "Server error"

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_invalid_rpc_method_rsvp_false(self, is_async):
        """Test no error response with unset rsvp"""
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )
        request = deepcopy(self.basic_request)
        request["rsvp"] = False

        @dispatcher.rpc_method
        def method(jarpc_request, param):
            return "2" + 2

        if is_async:
            response = await manager.get_response(json.dumps(request))
        else:
            response = manager.get_response(json.dumps(request))
        assert response is None

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_rpc_method_time_out(self, is_async, caplog):
        """Test rpc_method time out"""
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )
        request = deepcopy(self.basic_request)
        request["ts"] = datetime.now(tz=timezone.utc).timestamp()
        request["ttl"] = 1.0

        @dispatcher.rpc_method
        def method(jarpc_request, param):
            import time

            time.sleep(1)

        if is_async:
            response = await manager.get_response(json.dumps(request))
        else:
            response = manager.get_response(json.dumps(request))

        caplog.set_level(logging.WARNING)

        assert response is None
        assert "Request took too long to complete" in caplog.text

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_sync(self, is_async):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )

        @dispatcher.rpc_method
        def method(param):
            assert param == "value"
            return "return value"

        if is_async:
            response = await manager.get_response(json.dumps(self.basic_request))
        else:
            response = manager.get_response(json.dumps(self.basic_request))
        assert response is not None
        assert response.result == "return value"

    async def test_async(self):
        dispatcher = JarpcDispatcher()
        manager = AsyncJarpcManager(dispatcher)

        @dispatcher.rpc_method
        async def method(param):
            assert param == "value"
            return "return value"

        response = await manager.get_response(json.dumps(self.basic_request))
        assert response is not None
        assert response.result == "return value"

    async def test_task_cancellation(self):
        dispatcher = JarpcDispatcher()
        manager = AsyncJarpcManager(dispatcher)

        @dispatcher.rpc_method
        async def method(param):
            await asyncio.sleep(10)

        task = asyncio.ensure_future(
            manager.get_response(json.dumps(self.basic_request))
        )
        await asyncio.sleep(0.1)
        task.cancel()

        with pytest.raises(asyncio.CancelledError):
            await task


@pytest.mark.asyncio
class TestCallMethod:
    basic_request_kwargs = {
        "method": "method",
        "params": {"param": "value"},
        "request_id": "1",
        "ts": float(1 << 31),
        "ttl": 1.0,
        "rsvp": True,
    }

    @pytest.mark.parametrize("is_async", [False, True])
    @pytest.mark.parametrize(
        "params",
        [
            dict(a=1, b=2, c=3, x=0, y=1),
            dict(a=1, b=2, x=0),
            dict(a=1, b=2, x=0, y=1, z=2, abc="def"),
        ],
    )
    async def test_no_context(self, is_async, params):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )
        request_data = deepcopy(self.basic_request_kwargs)
        request_data["params"] = params
        request = JarpcRequest(**request_data)

        @dispatcher.rpc_method
        def method(a, b, c=0, *, x, y=1, **kw):
            return "return value"

        if is_async:
            result = await manager._call_method(method, request)
        else:
            result = manager._call_method(method, request)
        assert result == "return value"

    async def test_async(self):
        dispatcher = JarpcDispatcher()
        manager = AsyncJarpcManager(dispatcher)
        request_data = deepcopy(self.basic_request_kwargs)
        request = JarpcRequest(**request_data)

        @dispatcher.rpc_method
        async def method(jarpc_request, param):
            return "return value"

        result = await manager._call_method(method, request)
        assert result == "return value"

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_context(self, is_async):
        context = {"app": "some app"}
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher, context)
            if is_async
            else JarpcManager(dispatcher, context)
        )
        request = JarpcRequest(**self.basic_request_kwargs)

        @dispatcher.rpc_method
        def method(app, param):
            assert app == "some app"
            assert param == "value"
            return "return value"

        if is_async:
            result = await manager._call_method(method, request)
        else:
            result = manager._call_method(method, request)
        assert result == "return value"

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_request(self, is_async):
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher) if is_async else JarpcManager(dispatcher)
        )
        request = JarpcRequest(**self.basic_request_kwargs)

        @dispatcher.rpc_method
        def method(jarpc_request, param):
            assert jarpc_request is request
            assert param == "value"
            return "return value"

        if is_async:
            result = await manager._call_method(method, request)
        else:
            result = manager._call_method(method, request)
        assert result == "return value"

    @pytest.mark.parametrize("is_async", [False, True])
    async def test_context_conflict(self, is_async):
        context = {"app": "some app"}
        dispatcher = JarpcDispatcher()
        manager = (
            AsyncJarpcManager(dispatcher, context)
            if is_async
            else JarpcManager(dispatcher, context)
        )
        request_data = deepcopy(self.basic_request_kwargs)
        request_data["params"] = {"app": "my app", "param": "value"}
        request = JarpcRequest(**request_data)

        @dispatcher.rpc_method
        def method(app, param):
            assert False

        with pytest.raises(TypeError):
            if is_async:
                await manager._call_method(method, request)
            else:
                manager._call_method(method, request)



# -*- coding: utf-8 -*-
import json
import time
import uuid
from typing import Any, Generic, Optional, TypeVar

from pydantic import BaseModel, Field

from .errors import JarpcInvalidRequest, JarpcParseError, JarpcServerError

RequestT = TypeVar("RequestT")
ResponseT = TypeVar("ResponseT")


def json_dumps(data):
    """Default JSON serialiser."""
    return json.dumps(data, ensure_ascii=False)


def json_loads(data):
    """Default JSON deserialiser."""
    return json.loads(data)


class JarpcRequest(BaseModel, Generic[RequestT]):
    version: str = "1.0"
    method: str
    params: RequestT
    ts: float = Field(default_factory=time.time)
    ttl: float | None = None
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    rsvp: bool = True
    meta: dict[str, Any] | None = None

    def __repr__(self):
        return (
            f"<JarpcRequest version {self.version}, method {self.method}, params"
            f" {self.params}, ts {self.ts}, ttl {self.ttl}, id {self.id}, rsvp"
            f" {self.rsvp}, meta {self.meta}>"
        )

    @property
    def expired(self) -> bool:
        if self.ttl is None:
            return False
        return time.time() > self.ts + self.ttl

    #         except (TypeError, json.JSONDecodeError) as e:
    #             raise JarpcParseError(e) from e


class JarpcResponse(BaseModel, Generic[ResponseT]):
    result: ResponseT | None = None
    error: Any | None = None
    request_id: str | None = None
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    meta: dict[str, Any] | None = None

    def __repr__(self):
        return (
            f"<JarpcResponse id {self.id} result {self.result}, error {self.error},"
            f" request_id {self.request_id}, meta {self.meta}>"
        )

    @property
    def success(self) -> bool:
        return self.error is None



# -*- coding: utf-8 -*-
from .errors import JarpcMethodNotFound


class JarpcDispatcher:
    """Mapping for API methods. Effectively a dictionary wrapper."""

    def __init__(self, method_map: dict = None):
        if not isinstance(method_map, (dict, type(None))):
            raise TypeError
        self.method_map = method_map or dict()

    def __getitem__(self, item):
        try:
            return self.method_map[item]
        except KeyError as e:
            raise JarpcMethodNotFound(e) from e

    def rpc_method(self, f):
        """Decorator: adds `f` as RPC method.
        `f` can retrieve JarpcRequest object through optional `jarpc_request` argument.
        """
        self.method_map[f.__name__] = f
        return f

    def declare_method(self, name: str | None = None):
        """Decorator: adds `f` as RPC method.
        `f` can retrieve JarpcRequest object through optional `jarpc_request` argument.
        """

        def decorated(f):
            self.method_map[(name.__str__() if name else None) or f.__name__] = f
            return f

        return decorated

    def add_rpc_method(self, f, name=None):
        """Adds `f` as RPC method.
        If `name` is not None, it is used as method name.
        `f` can retrieve JarpcRequest object through optional `jarpc_request` argument.
        """
        self.method_map[name or f.__name__] = f

    def update(self, dispatcher):
        """Add methods from `dispatcher`, overriding on any collisions."""
        self.method_map.update(dispatcher.method_map)



import inspect
from inspect import _empty, Signature, Parameter
from typing import Any, Type, Callable, Generator

from pydantic import BaseModel, create_model

from jarpcdantic import AsyncJarpcClient, JarpcClient
from jarpcdantic.utils import process_return_value


class JarpcClientRouter:
    def __init__(
        self,
        prefix: str | None = None,
        client: AsyncJarpcClient | JarpcClient | None = None,
        is_absolute_prefix: bool = False,
    ):
        """
       Initializes the JARPC client router.

       :param prefix: Optional prefix for the router.
       :type prefix: str
       :param client: Optional client instance (AsyncJarpcClient or JarpcClient).
       :type client: Union[AsyncJarpcClient, JarpcClient]
       :param is_absolute_prefix: Whether the prefix is absolute.
       :type is_absolute_prefix: bool
       """
        self._client: AsyncJarpcClient | JarpcClient | None = client
        self._prefix: str | None = prefix
        self._is_absolute_prefix: bool = is_absolute_prefix
        self._method_map: dict[str, Callable[..., Any]] = {}

        self._decorate_endpoints()

    def _filter_attributes(self) -> Generator[tuple[str, Any], None, None]:
        """
        Filters and yields non-private attributes of the class.

        :yield: Tuples of attribute names and values.
        """
        for attr_name, attr_value in self.__class__.__dict__.items():
            if attr_name.startswith("_") or isinstance(attr_value, property):
                continue
            yield attr_name, attr_value

    @staticmethod
    def _is_nested_router(attr_value: Any) -> bool:
        """
        Checks if the attribute is an instance of JarpcClientRouter.

        :param attr_value: Attribute value to check.
        :type attr_value: Any
        :return: True if the attribute is a nested router, False otherwise.
        """
        return isinstance(attr_value, JarpcClientRouter)

    @staticmethod
    def _is_not_processed_endpoint(attr_name: str, attr_value: Any, is_nested: bool) -> bool:
        """
        Checks if the attribute is not a processed endpoint.

        :param attr_name: Name of the attribute.
        :type attr_name: str
        :param attr_value: Value of the attribute.
        :type attr_value: Any
        :param is_nested: Whether the router is nested.
        :return: True if the attribute is an endpoint, False otherwise.
        """
        return (
            not is_nested
            and not attr_name.startswith("_")
            and hasattr(attr_value, "__annotations__")
            and "return" in attr_value.__annotations__
        )

    def _process_attributes(self, is_nested: bool = False) -> None:
        """
        Handles attributes, processing nested routers and endpoints.

        :param is_nested: Whether the router is nested.
        :type is_nested: bool
        """
        for attr_name, attr_value in self._filter_attributes():
            if self._is_nested_router(attr_value):
                self._proceed_nested_router(attr_name, attr_value)
            elif self._is_not_processed_endpoint(attr_name, attr_value, is_nested):
                self._proceed_endpoint(attr_name, attr_value)

    def _decorate_endpoints(self, is_nested: bool = False) -> None:
        """
        Decorates methods to handle attributes.

        :param is_nested: Whether the router is nested.
        :type is_nested: bool
        """
        self._process_attributes(is_nested)

    @staticmethod
    def _filter_parameters(attr_signature: Signature) -> dict[str, Parameter]:
        """
        Filters parameters to exclude 'self', 'args', 'kwargs', and private attributes.

        :param attr_signature: Signature of the attribute.
        :type attr_signature: Signature
        :return: Filtered parameters as a dictionary.
        """
        return  {
            k: v
            for k, v in attr_signature.parameters.items()
            if k not in ["self", "args", "kwargs"] and not k.startswith("_")
        }

    @staticmethod
    def _determine_request_model(filtered_parameters: dict[str, inspect.Parameter], attr_signature: Signature) -> Type[BaseModel] | None:
        """
        Determines the request model for the endpoint.

        :param filtered_parameters: Filtered parameters of the attribute.
        :type filtered_parameters: dict
        :param attr_signature: Signature of the attribute.
        :type attr_signature: Signature
        :return: Request model if applicable, None otherwise.
        """
        request_model: Type[BaseModel] | None = None

        if "_model" in attr_signature.parameters:
            request_model: Type[BaseModel] = attr_signature.parameters["_model"].default
        # elif (
        #     len(filtered_parameters) == 1
        #     and isinstance(
        #         filtered_parameters[list(filtered_parameters.keys())[0]].annotation, type
        #     )
        #     and issubclass(
        #         filtered_parameters[list(filtered_parameters.keys())[0]].annotation, BaseModel
        #     )
        # ):
        #     parameter = filtered_parameters[list(filtered_parameters.keys())[0]]
        #     request_model = (
        #         parameter.annotation if parameter.annotation is not _empty else None
        #     )
        elif len(filtered_parameters) >= 1:
            request_model = create_model(
                "DynamicModel",
                **{
                    k: (
                        Any if v.annotation is _empty else v.annotation,
                        ... if v.default is _empty else v.default,
                    )
                    for k, v in filtered_parameters.items()
                },
            )

        return request_model

    def _wrap_endpoint(self, endpoint_name: str, request_model: Type[BaseModel], endpoint_signature: Signature) -> Any:
        """
        Wraps an endpoint method to handle parameter validation and sending requests via the client.

        :param endpoint_name: Name of the endpoint.
        :type endpoint_name: str
        :param request_model: Request model used for validation, if applicable.
        :type request_model: Type[BaseModel]
        :param endpoint_signature: Signature of the original method.
        :type endpoint_signature: Signature
        :return: Wrapped function to be used as endpoint.
        """
        wrapped_attr = self._wrap(self, endpoint_name, request_model, endpoint_signature)
        wrapped_attr.__annotations__ = {
            param_name: param.annotation for param_name, param in endpoint_signature.parameters.items()
        }
        wrapped_attr.__annotations__['return'] = endpoint_signature.return_annotation
        return wrapped_attr

    def _proceed_endpoint(self, endpoint_name: str, endpoint_method: Any) -> None:
        """
        Processes an endpoint by wrapping it and adding it to the method map.

        :param endpoint_name: Name of the endpoint.
        :type endpoint_name: str
        :param endpoint_method: Original method of the endpoint.
        :type endpoint_method: Any
        """
        endpoint_signature = inspect.signature(endpoint_method)
        filtered_parameters = self._filter_parameters(endpoint_signature)
        request_model = self._determine_request_model(filtered_parameters, endpoint_signature)
        wrapped_endpoint = self._wrap_endpoint(endpoint_name, request_model, endpoint_signature)

        self._method_map[endpoint_name] = wrapped_endpoint
        setattr(self, endpoint_name, wrapped_endpoint)

    def _proceed_nested_router(
        self, attr_router_name: str, nested_router: "JarpcClientRouter"
    ) -> None:
        """
        Processes a nested router by setting prefixes, clients, and decorating its methods.

        :param attr_router_name: Name of the nested router attribute.
        :type attr_router_name: str
        :param nested_router: Instance of the nested router.
        :type nested_router: JarpcClientRouter
        """
        if nested_router._prefix is None:
            nested_router._prefix = attr_router_name

        if nested_router._is_absolute_prefix:
            return

        if self._prefix and not nested_router._prefix.startswith(self._prefix):
            nested_router._prefix = f"{self._prefix}.{nested_router._prefix}".strip(".")

        if not nested_router._client and self._client:
            nested_router._client = self._client

        nested_router._decorate_endpoints(is_nested=True)

    def set_client(self, client: AsyncJarpcClient | JarpcClient) -> None:
        """
        Sets the client for this router and all nested routers.

        :param client: JARPC Client instance
        :type client: AsyncJarpcClient | JarpcClient
        """
        self._client = client
        for attr_name, attr_value in self.__class__.__dict__.items():
            if attr_name.startswith("_") or isinstance(attr_value, property):
                continue

            if isinstance(attr_value, JarpcClientRouter):
                attr_value.set_client(client)

    @staticmethod
    def _wrap(
        instance: "JarpcClientRouter",
        method_name: str,
        request_model: Type[BaseModel] | None,
        method_signature: Signature,
    ) -> Callable[..., Any]:
        """
        Wraps an endpoint method to handle parameter validation and sending requests via the client.

        :param instance: The instance of the router.
        :type instance: JarpcClientRouter
        :param method_name: The name of the endpoint method.
        :type method_name: str
        :param request_model: The request model used for validation, if applicable.
        :type request_model: Type[BaseModel] | None
        :param method_signature: The signature of the original method.
        :type method_signature: Signature
        :return: A wrapped asynchronous function that handles parameter validation and makes a request.
        """

        async def wrapped(*args: Any, **kwargs: Any) -> Any:
            # Separate parameters for service (prefixed with "_") and clear parameters
            service_kwargs = {k: v for k, v in kwargs.items() if k.startswith("_")}
            clear_kwargs = {k: v for k, v in kwargs.items() if not k.startswith("_")}

            # If a request model is defined, validate and construct request parameters
            if request_model is not None:
                # Initialize parameters with default values from the signature
                args_with_defaults = {
                    param_name: param.default
                    for param_name, param in method_signature.parameters.items()
                }

                # Update default arguments with positional arguments
                for index, value in enumerate(args):
                    param_name = list(method_signature.parameters.keys())[
                        index + 1
                    ]  # Skip "self"
                    args_with_defaults[param_name] = value

                # Merge keyword arguments
                combined_params = {**args_with_defaults, **clear_kwargs}

                # Validate and parse parameters using the model if it's a subclass of BaseModel
                if issubclass(request_model, BaseModel):
                    params = request_model(**combined_params)
                else:
                    # Filter out parameters with no default value or empty value
                    params = {
                        k: v for k, v in combined_params.items() if v is not _empty
                    }
            else:
                # If no model is defined, pass only the clear parameters
                params = {}

            # Make a request using the client, if defined
            if instance._client is not None:
                # Construct the full method name with prefix, if any
                full_method_name = (
                    f"{instance._prefix}.{method_name}" if instance._prefix else method_name
                )

                for service_key in {"ts", "ttl", "request_id", "rsvp", "durable"}:
                    if f"_{service_key}" in service_kwargs:
                        service_kwargs[service_key] = service_kwargs[f"_{service_key}"]
                        del service_kwargs[f"_{service_key}"]

                # Send the request and process the response
                response = await instance._client(
                    method=full_method_name,
                    params=params,
                    generic_request_type=request_model,
                    generic_response_type=method_signature.return_annotation,
                    **service_kwargs,
                )
                return process_return_value(method_signature.return_annotation, response)
            else:
                # If no client is set, print the call for debugging
                full_method_name = (
                    f"{instance._prefix}.{method_name}" if instance._prefix else method_name
                )
                print("Client is None, call:", full_method_name, params)

        # Provide string representation of the wrapped method name
        def __str__(*args, **kwargs) -> str:
            return f"{instance._prefix}.{method_name}" if instance._prefix else method_name

        # Attach string representation to the wrapped function
        wrapped.__str__ = __str__
        wrapped.__repr__ = __str__

        return wrapped



# -*- coding: utf-8 -*-
"""
JARPC exception classes.

Modelled after JSON-RPC 2.0 errors.

"""
import inspect
import sys


class JarpcError(Exception):
    """Base JARPC exception"""

    code = None
    message = None

    def __init__(self, data=None):
        if isinstance(data, Exception):
            self.data = f"{data.__class__.__name__}: {data}"
        else:
            self.data = data

    def __str__(self):
        return f"{self.code} {self.message}: {self.data}"

    def as_dict(self):
        return {"code": self.code, "message": self.message, "data": self.data}


class JarpcParseError(JarpcError):
    """Parse Error: invalid JSON format."""

    code = -32700
    message = "Parse error"


class JarpcInvalidRequest(JarpcError):
    """Invalid Request: the JSON object is not a JARPC 1.0 request."""

    code = -32600
    message = "Invalid Request"


class JarpcMethodNotFound(JarpcError):
    """Method not found: method with requested name does not exist."""

    code = -32601
    message = "Method not found"


class JarpcInvalidParams(JarpcError):
    """Invalid params: invalid method call."""

    code = -32602
    message = "Invalid params"


class JarpcInternalError(JarpcError):
    """Internal error: internal JARPC error."""

    code = -32603
    message = "Internal error"


class JarpcTimeout(JarpcError):
    """Timeout: request could not be completed in time."""

    code = -32604
    message = "Timeout"


class JarpcServerError(JarpcError):
    """Server error: server could not complete the request."""

    code = -32000
    message = "Server error"


class JarpcUnknownError(JarpcError):
    """Unknown error: unknown exception code"""

    def __init__(self, code, message, data):
        self.code = code
        self.message = message
        super().__init__(data=data)


# Should be thrown in dispatch methods.

# 1xxx - ошибки доступа


class JarpcUnauthorized(JarpcError):
    """Unauthorized: Similar to JarpcForbidden, but specifically for use when authentication is required and
    has failed or has not yet been provided."""

    code = 1000
    message = "Unauthorized"


class JarpcForbidden(JarpcError):
    """Forbidden: The request was valid, but the server is refusing action.
    The user might not have the necessary permissions for a resource, or may need an account of some sort.
    """

    code = 1001
    message = "Forbidden"


# 2xxx - неправильные входные данные


class JarpcValidationError(JarpcError):
    """ValidationError: error validating input parameters"""

    code = 2000
    message = "Validation error"


# 3xxx - ошибки интеграций


class JarpcExternalServiceUnavailable(JarpcError):
    """ExternalServiceUnavailable: ошибка запроса удаленной системы"""

    code = 3000
    message = "External service unavailable"


def _get_exceptions():
    return [
        exception
        for _, exception in inspect.getmembers(sys.modules[__name__], inspect.isclass)
        if issubclass(exception, JarpcError)
        and exception not in {JarpcError, JarpcUnknownError}
    ]


_exception_codes = {exception.code: exception for exception in _get_exceptions()}


def raise_exception(code, data=None, message=None):
    """Recreate JARPC exception from JARPC error code."""
    try:
        exception = _exception_codes[code](data=data)
    except KeyError:
        exception = JarpcUnknownError(code=code, message=message, data=data)

    raise exception



# -*- coding: utf-8 -*-
import json
from unittest import mock

import pytest
from freezegun import freeze_time

from jarpcdantic import (
    AsyncJarpcClient,
    JarpcClient,
    JarpcRequest,
    JarpcServerError,
    JarpcTimeout,
    JarpcValidationError,
)


class TestJarpcClient:
    @pytest.mark.parametrize("is_async", [False, True])
    @freeze_time("2012-01-14")
    def test_prepare_request(self, is_async):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(transport=None)

        with mock.patch("uuid.uuid4") as uuid_mock:
            uuid_mock.return_value = "example-uuid"
            request = jarpc_client._prepare_request(
                method="test_method", params={"a": 1}
            )

        assert request.model_dump() == {
            "method": "test_method",
            "params": {"a": 1},
            "id": "example-uuid",
            "version": "1.0",
            "ts": 1326499200.0,
            "ttl": None,
            "rsvp": True,
        }

    @pytest.mark.parametrize(
        "rsvp, ttl, default_ttl, default_rpc_ttl, default_notification_ttl,"
        " expected_ttl",
        [
            (True, None, None, None, None, None),
            (False, None, None, None, None, None),
            (True, 10.0, None, None, None, 10.0),
            (False, 10.0, None, None, None, 10.0),
            (True, None, 10.0, None, None, 10.0),
            (False, None, 10.0, None, None, 10.0),
            (True, 30.0, 10.0, None, None, 30.0),
            (False, 30.0, 10.0, None, None, 30.0),
            (True, None, None, 10.0, 20.0, 10.0),
            (False, None, None, 10.0, 20.0, 20.0),
            (True, None, 30.0, 10.0, 20.0, 10.0),
            (False, None, 30.0, 10.0, 20.0, 20.0),
            (True, 30.0, None, 10.0, 20.0, 30.0),
            (False, 30.0, None, 10.0, 20.0, 30.0),
            (True, 40.0, 30.0, 10.0, 20.0, 40.0),
            (False, 40.0, 30.0, 10.0, 20.0, 40.0),
        ],
    )
    @pytest.mark.parametrize("is_async", [False, True])
    @freeze_time("2012-01-14")
    def test_prepare_request_default_ttl(
        self,
        is_async,
        rsvp,
        ttl,
        default_ttl,
        default_rpc_ttl,
        default_notification_ttl,
        expected_ttl,
    ):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(
            transport=None,
            default_ttl=default_ttl,
            default_rpc_ttl=default_rpc_ttl,
            default_notification_ttl=default_notification_ttl,
        )

        with mock.patch("uuid.uuid4") as uuid_mock:
            uuid_mock.return_value = "example-uuid"
            request = jarpc_client._prepare_request(
                method="test_method", params={"a": 1}, ttl=ttl, rsvp=rsvp
            )

        assert request.model_dump() == {
            "method": "test_method",
            "params": {"a": 1},
            "id": "example-uuid",
            "version": "1.0",
            "ts": 1326499200.0,
            "ttl": expected_ttl,
            "rsvp": rsvp,
        }

    @pytest.mark.parametrize("is_async", [False, True])
    @freeze_time("2012-01-14")
    def test_prepare_request_durable(self, is_async):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(
            transport=None,
            default_ttl=1.0,
            default_rpc_ttl=2.0,
            default_notification_ttl=3.0,
        )

        with mock.patch("uuid.uuid4") as uuid_mock:
            uuid_mock.return_value = "example-uuid"
            request = jarpc_client._prepare_request(
                method="test_method", params={"a": 1}, ttl=4.0, durable=True
            )

        assert request.model_dump() == {
            "method": "test_method",
            "params": {"a": 1},
            "id": "example-uuid",
            "version": "1.0",
            "ts": 1326499200.0,
            "ttl": None,
            "rsvp": True,
        }

    @pytest.mark.parametrize("is_async", [False, True])
    def test_parse_response_no_rsvp(self, is_async):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(transport=None)
        assert jarpc_client._parse_response(response_string="", rsvp=False) is None

    @pytest.mark.parametrize("is_async", [False, True])
    def test_parse_response_ok(self, is_async):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(transport=None)
        result = {
            "result": "some result",
            "request_id": "123-456-788",
            "id": "123-456-789",
        }
        assert (
            jarpc_client._parse_response(response_string=json.dumps(result), rsvp=True)
            == result["result"]
        )

    @pytest.mark.parametrize("is_async", [False, True])
    def test_parse_response_error(self, is_async):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(transport=None)
        error = {
            "error": {"code": 2000, "message": "Validation error", "data": "some data"},
            "request_id": "123-456-788",
            "id": "123-456-789",
        }
        with pytest.raises(JarpcValidationError) as e:
            jarpc_client._parse_response(response_string=json.dumps(error), rsvp=True)
        assert e.value.code == error["error"]["code"]
        assert e.value.message == error["error"]["message"]
        assert e.value.data == error["error"]["data"]

    @pytest.mark.parametrize("is_async", [False, True])
    def test_parse_response_invalid_json(self, is_async):
        client = AsyncJarpcClient if is_async else JarpcClient
        jarpc_client = client(transport=None)

        with pytest.raises(JarpcServerError):
            jarpc_client._parse_response(response_string="", rsvp=True)

    @pytest.mark.parametrize("is_async", [False, True])
    @pytest.mark.asyncio
    @freeze_time("2012-01-14")
    async def test_call(self, is_async):
        request_id = "123-456-788"
        response = {
            "result": "some result",
            "request_id": request_id,
            "id": "123-456-789",
        }
        transport = lambda *args, **kwargs: json.dumps(response)
        jarpc_client = JarpcClient(transport=transport)

        call_kwargs = dict(
            method="method", params={"param1": 1}, id=request_id, extra_kwarg=1
        )
        call_result = jarpc_client(**call_kwargs)
        assert call_result == response["result"]

    @pytest.mark.parametrize("is_async", [False, True])
    @pytest.mark.asyncio
    @freeze_time("2012-01-14")
    async def test_call_simple_syntax(self, is_async):
        response = {
            "result": "some result",
            "request_id": "no check",
            "id": "123-456-789",
        }

        transport = lambda *args, **kwargs: json.dumps(response)
        jarpc_client = JarpcClient(transport=transport)

        call_result = jarpc_client.method(param1=1)

        assert call_result == response["result"]



# -*- coding: utf-8 -*-
import json
from datetime import datetime

import pytest
from freezegun import freeze_time

from jarpcdantic import (
    JarpcInvalidRequest,
    JarpcParseError,
    JarpcRequest,
    JarpcResponse,
    JarpcServerError,
)

VALID_REQUEST_KWARGS = {
    "method": "get_result",
    "params": {"a": 1, "b": 2},
    "ts": 0.0,
    "ttl": 10.0,
    "id": "429284302",
    "rsvp": True,
}
VALID_REQUEST_DATA = {**VALID_REQUEST_KWARGS, "version": "1.0"}


class TestJarpcRequest:
    @freeze_time("1996-06-06 06:06:06")
    def test_expired(self):
        jarpc_request = JarpcRequest(**VALID_REQUEST_KWARGS)
        jarpc_request.ttl = 1
        jarpc_request.ts = datetime(1996, 6, 6, 6, 6, 6).timestamp() - 2
        assert jarpc_request.expired

    def test_expired_ttl_empty(self):
        jarpc_request = JarpcRequest(**VALID_REQUEST_KWARGS)
        jarpc_request.ttl = None
        assert not jarpc_request.expired

    def test_data(self):
        jarpc_request = JarpcRequest(**VALID_REQUEST_KWARGS)
        assert jarpc_request.model_dump() == VALID_REQUEST_DATA

    def test_serialize(self):
        jarpc_request = JarpcRequest(**VALID_REQUEST_KWARGS)
        assert json.loads(jarpc_request.model_dump_json()) == VALID_REQUEST_DATA

    def test_from_json(self):
        jarpc_request = JarpcRequest.model_validate_json(json.dumps(VALID_REQUEST_DATA))
        assert jarpc_request.model_dump() == VALID_REQUEST_DATA

    def test_from_data(self):
        jarpc_request = JarpcRequest.model_validate(VALID_REQUEST_DATA)
        assert jarpc_request.method == VALID_REQUEST_DATA["method"]
        assert jarpc_request.version == VALID_REQUEST_DATA["version"]
        assert jarpc_request.params == VALID_REQUEST_DATA["params"]
        assert jarpc_request.ts == VALID_REQUEST_DATA["ts"]
        assert jarpc_request.ttl == VALID_REQUEST_DATA["ttl"]
        assert jarpc_request.id == VALID_REQUEST_DATA["id"]
        assert jarpc_request.rsvp == VALID_REQUEST_DATA["rsvp"]


VALID_RESULT = {
    "error": None,
    "result": "some result",
    "request_id": "123-456-788",
    "id": "123-456-789",
}

VALID_ERROR = {
    "result": None,
    "error": {"code": 2000, "message": "Validation error", "data": "some data"},
    "request_id": "123-456-788",
    "id": "123-456-789",
}


class TestJarpcResponse:
    @pytest.mark.parametrize("data", [VALID_RESULT, VALID_ERROR])
    def test_data(self, data):
        jarpc_response = JarpcResponse(**data)
        assert jarpc_response.model_dump() == data

    @pytest.mark.parametrize("data", [VALID_RESULT, VALID_ERROR])
    def test_serialize(self, data):
        jarpc_response = JarpcResponse(**data)
        assert json.loads(jarpc_response.model_dump_json()) == data

    @pytest.mark.parametrize("data", [VALID_RESULT, VALID_ERROR])
    def test_from_json(self, data):
        jarpc_response = JarpcResponse.model_validate_json(json.dumps(data))
        assert jarpc_response.model_dump() == data

    @pytest.mark.parametrize("data", [VALID_RESULT, VALID_ERROR])
    def test_from_data(self, data):
        jarpc_response = JarpcResponse.model_validate(data)
        assert jarpc_response.result == data.get("result")
        assert jarpc_response.error == data.get("error")
        assert jarpc_response.request_id == data["request_id"]
        assert jarpc_response.id == data["id"]



# -*- coding: utf-8 -*-
import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def get_version():
    """Get version from the package without actually importing it."""
    init = read("jarpcdantic/__init__.py")
    for line in init.split("\n"):
        if line.startswith("__version__"):
            return eval(line.split("=")[1])


setup(
    name="jarpcdantic",
    version=get_version(),
    description="JSON Advanced RPC with Pydantic",
    packages=["jarpcdantic"],
    requires=["pydantic"],
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    maintainer="WhiteApfel",
    maintainer_email="white@pfel.ru",
    url='https://github.com/whiteapfel/jarpcdantic/',
    download_url='https://pypi.org/project/jarpcdantic/',
    license="Mozilla Public License 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.12",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Topic :: Utilities",
    ],
)



# -*- coding: utf-8 -*-
import inspect
import logging
from asyncio import CancelledError
from collections import deque
from typing import Iterable, Optional, get_origin, Union, get_args

from pydantic import BaseModel
from pydantic_core import ValidationError

from .context import meta_context_var
from .dispatcher import JarpcDispatcher
from .errors import JarpcError, JarpcInvalidParams, JarpcParseError, JarpcServerError
from .format import JarpcRequest, JarpcResponse, json_dumps, json_loads
from .utils import convert_params_to_models, process_return_value

logger = logging.getLogger(__name__)


def get_args_representation(args: Iterable) -> str:
    """
    ['c', 'a', 'b'] -> "a, b, c"
    {'b',"a"} -> "a, b"
    """
    return ", ".join(sorted(args))


def prepare_context_params(method, request, context):
    """
    Подготавливает параметры контекста для вызова метода.
    """
    context_params = {}
    method_sig = inspect.signature(method)
    if "jarpc_request" in method_sig.parameters:
        context_params["jarpc_request"] = request
    for param, value in context.items():
        if param in method_sig.parameters:
            context_params[param] = value
    return context_params, method_sig


def check_function_call(fun, kwargs: dict, context: dict) -> (bool, Optional[str]):
    """
    Check that `kwargs` match signature of `fun` given `context`.
    Positional arguments in call are not supported.

    Minimal valid argument set: positional args with no default + kw-only args with no default,
                                except for what is provided by context.
    Maximum valid argument set: all positional + all kw-only args IF no **varkw present,
                                except for what is provided by context.
    If **varkw is present, there is no upper limit.
    *varargs has no effect because there are no positional args in call.
    If fun is a method, args[0] is self. If fun is a callable object, we have to look at its __call__

    :returns (is_ok, explanation)
    """
    argspec = inspect.getfullargspec(fun)
    args_deque = deque(argspec.args)
    if inspect.ismethod(fun) or inspect.ismethod(fun.__call__):
        # skip self
        args_deque.popleft()
    allowed_args = set(args_deque)
    if argspec.defaults:
        # skip defaults
        for _ in argspec.defaults:
            args_deque.pop()
    required_args = set(args_deque)

    required = (
        (required_args | set(argspec.kwonlyargs))
        - (argspec.kwonlydefaults or {}).keys()
        - context.keys()
        - {"jarpc_request"}
    )
    if required - kwargs.keys():
        return (
            False,
            f"Missing arguments: {get_args_representation(required - kwargs.keys())}",
        )

    if argspec.varkw is None:
        allowed = (
            (allowed_args | set(argspec.kwonlyargs))
            - context.keys()
            - {"jarpc_request"}
        )
        if kwargs.keys() - allowed:
            return (
                False,
                (
                    "Unexpected arguments:"
                    f" {get_args_representation(kwargs.keys() - allowed)}"
                ),
            )
    else:
        # if **varkw is present, anything is considered allowed except for args from context
        restricted_intersection = kwargs.keys() & (context.keys() | {"jarpc_request"})
        if restricted_intersection:
            return (
                False,
                (
                    "Unavailable arguments:"
                    f" {get_args_representation(restricted_intersection)}"
                ),
            )
    return True, None


class JarpcManager:
    def __init__(
        self,
        dispatcher: JarpcDispatcher,
        context: dict = None,
        loads=json_loads,
        dumps=json_dumps,
    ):
        self.dispatcher = dispatcher
        self.context = (
            context or dict()
        )  # per-manager context cannot contain jarpc_request
        self.loads = loads
        self.dumps = dumps

    def handle(self, request: str) -> Optional[str]:
        """Handle request string, producing either response string or None if no response is required."""
        jarpc_response = self.get_response(request_string=request)
        if jarpc_response is not None:
            return jarpc_response.model_dump_json()

    def get_response(self, request_string: str) -> Optional[JarpcResponse]:
        """Returns either JarpcResponse or None if no response is required."""
        request_id = None
        rsvp = True
        context_token = None
        try:
            try:
                request = JarpcRequest.model_validate_json(request_string)
            except ValidationError:
                raise JarpcParseError()
            if request.expired:
                logger.warning(f"Request arrived too late: {request}")
                return None

            request_id = request.id
            rsvp = request.rsvp
            meta = request.meta
            context_token = meta_context_var.set(meta)

            method = self.dispatcher[request.method]
            try:
                result = self._call_method(method, request)
            except TypeError:
                is_call_ok, explanation = check_function_call(
                    method, request.params, self.context
                )
                if is_call_ok:
                    raise
                logger.debug(
                    f"wrong signature in call to {request.method}: {explanation}"
                )
                raise JarpcInvalidParams(explanation)

            if request.expired:
                logger.warning(f"Request took too long to complete: {request}")
                return None
            return JarpcResponse(request_id=request_id, result=result) if rsvp else None
        except JarpcError as e:
            logger.debug(e, exc_info=True)
            return (
                JarpcResponse(request_id=request_id, error=e.as_dict())
                if rsvp
                else None
            )
        except Exception as e:
            logger.exception(e)
            return (
                JarpcResponse(
                    request_id=request_id, error=JarpcServerError(e).as_dict()
                )
                if rsvp
                else None
            )
        finally:
            if context_token is not None:
                meta_context_var.reset(context_token)

    def _call_method(self, method, request):
        """
        Универсальный вызов метода с обработкой результата (синхронная версия).
        """
        context_params, method_sig = prepare_context_params(
            method, request, self.context
        )
        converted_params = convert_params_to_models(request.params, method_sig)
        final_params = {**converted_params, **context_params}
        result = method(**final_params)
        return_annotation = method_sig.return_annotation
        return process_return_value(return_annotation, result)


class AsyncJarpcManager(JarpcManager):
    async def handle(self, request: str) -> Optional[str]:
        """Handle request string, producing either response string or None if no response is required."""
        jarpc_response = await self.get_response(request_string=request)
        if jarpc_response is not None:
            return jarpc_response.model_dump_json()

    async def get_response(self, request_string: str) -> Optional[JarpcResponse]:
        """Returns either JarpcResponse or None if no response is required."""
        request_id = None
        rsvp = True
        context_token = None
        try:
            try:
                request = JarpcRequest.model_validate_json(request_string)
            except ValidationError:
                raise JarpcParseError()
            if request.expired:
                logger.warning(f"Request arrived too late: {request}")
                return None

            request_id = request.id
            rsvp = request.rsvp
            meta = request.meta
            context_token = meta_context_var.set(meta)

            method = self.dispatcher[request.method]
            try:
                result = await self._call_method(method, request)
            except TypeError:
                is_call_ok, explanation = check_function_call(
                    method, request.params, self.context
                )
                if is_call_ok:
                    raise
                logger.debug(
                    f"wrong signature in call to {request.method}: {explanation}"
                )
                raise JarpcInvalidParams(explanation)

            if request.expired:
                logger.warning(f"Request took too long to complete: {request}")
                return None
            return JarpcResponse(request_id=request_id, result=result) if rsvp else None
        except CancelledError:
            raise
        except JarpcError as e:
            logger.debug(e, exc_info=True)
            return (
                JarpcResponse(request_id=request_id, error=e.as_dict())
                if rsvp
                else None
            )
        except Exception as e:
            logger.exception(e)
            return (
                JarpcResponse(
                    request_id=request_id, error=JarpcServerError(e).as_dict()
                )
                if rsvp
                else None
            )
        finally:
            if context_token is not None:
                meta_context_var.reset(context_token)

    async def _call_method(self, method, request):
        """
        Универсальный вызов метода с обработкой результата (асинхронная версия).
        """
        context_params, method_sig = prepare_context_params(
            method, request, self.context
        )
        converted_params = convert_params_to_models(request.params, method_sig)
        final_params = {**converted_params, **context_params}
        result = method(**final_params)
        if inspect.isawaitable(result):
            result = await result
        return_annotation = method_sig.return_annotation
        return process_return_value(return_annotation, result)
